package com.maximussoft.socketssl;

import android.util.Log;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.BA.ShortName;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.TimeUnit;

@ShortName("AsyncStreamsSSL")

public class AsyncStreamsSSL {
	private String eventName;
	private BA ba;
	private AIN ain;
	private Thread tin;
	private AOUT aout;
	private Thread tout;

	public void Initialize(BA ba, InputStream In, OutputStream Out,
			String EventName) throws IOException {
		shared(ba, In, Out, EventName, false, false);
	}

	public void InitializePrefix(BA ba, InputStream In, boolean BigEndian,
			OutputStream Out, String EventName) throws IOException {
		shared(ba, In, Out, EventName, BigEndian, true);
	}

	private void shared(BA ba, InputStream In, OutputStream Out,
			String EventName, boolean BigEndian, boolean Prefix)
			throws IOException {
		if (IsInitialized())
			Close();
		this.ba = ba;
		this.eventName = EventName.toLowerCase(BA.cul);
		if (In != null) {
			this.ain = new AIN(In, BigEndian, Prefix);
			this.tin = new Thread(this.ain);
			this.tin.setDaemon(true);
			this.tin.start();
		}
		if (Out != null) {
			this.aout = new AOUT(Out, BigEndian, Prefix);
			this.tout = new Thread(this.aout);
			this.tout.setDaemon(true);
			this.tout.start();
		}
	}

	public boolean IsInitialized() {
		return (this.ain != null) || (this.aout != null);
	}

	public boolean Write(byte[] Buffer) {
		return Write2(Buffer, 0, Buffer.length);
	}

	public boolean Write2(byte[] Buffer, int Start, int Length) {
		//return this.aout.put(Buffer, Start, Length);
		boolean ok = this.aout.put(Buffer, Start, Length);
		try {
			this.aout.out.write("\n".getBytes());
		} catch (IOException e) {
			//e.printStackTrace();
			Log.w("B4A", "** AsyncStreamsSSL write error **");
		}
		Log.w("B4A", "** AsyncStreamsSSL write done **");
		return ok;
	}

	public int getOutputQueueSize() {
		return this.aout.queue.size();
	}

	public void Close() throws IOException {
		if ((this.tin != null) && (this.ain != null)) {
			this.ain.close();
			if (Thread.currentThread() != this.tin)
				this.tin.interrupt();
		}
		if ((this.tout != null) && (this.aout != null)) {
			this.aout.close();
			if (Thread.currentThread() != this.tout)
				this.tout.interrupt();
		}
		this.ain = null;
		this.aout = null;
	}

	private class AIN implements Runnable {
		private final InputStream in;
		private byte[] buffer = new byte[8192];
		private final boolean prefix;
		private volatile boolean working = true;
		private String ev = AsyncStreamsSSL.this.eventName + "_newdata";
		private ByteBuffer bb;

		public AIN(InputStream in, boolean bigEndian, boolean prefix) {
			this.in = in;
			this.prefix = prefix;
			if (prefix) {
				this.bb = ByteBuffer.wrap(new byte[4]);
				this.bb.order(bigEndian ? ByteOrder.BIG_ENDIAN
						: ByteOrder.LITTLE_ENDIAN);
			}
		}

		public void run() {
			try {
				int count = 0;

				while (this.working) {
					byte[] data;
					if (!this.prefix) {
						count = this.in.read(this.buffer);
						if (count == 0)
							continue;
						if (count < 0) {
							closeUnexpected();
							break;
						}
						if (!this.working)
							break;
						data = new byte[count];
						System.arraycopy(this.buffer, 0, data, 0, count);
					} else {
						do {
							int c = this.in.read(this.buffer, count,
									this.buffer.length - count);
							if (c == -1) {
								closeUnexpected();
								break;
							}
							count += c;
						} while (count < 4);

						if (!this.working)
							break;
						this.bb.clear();
						this.bb.put(this.buffer, 0, 4);
						int msgLength = this.bb.getInt(0);
						if (msgLength > this.buffer.length) {
							byte[] newbuffer = new byte[msgLength + 4];
							System.arraycopy(this.buffer, 0, newbuffer, 0,
									count);
							this.buffer = newbuffer;
						}
						while (count < msgLength + 4) {
							int c = this.in.read(this.buffer, count,
									this.buffer.length - count);
							if (c == -1) {
								closeUnexpected();
								break;
							}
							count += c;
						}
						data = new byte[msgLength];
						System.arraycopy(this.buffer, 4, data, 0, data.length);
						if (count > msgLength + 4) {
							int nextSegment = count - (msgLength + 4);
							System.arraycopy(this.buffer, msgLength + 4,
									this.buffer, 0, nextSegment);
							count = nextSegment;
						} else {
							count = 0;
						}
					}

					AsyncStreamsSSL.this.ba.raiseEventFromDifferentThread(
							AsyncStreamsSSL.this, null, 0, this.ev, true,
							new Object[] { data });
				}
			} catch (Exception e) {
				if (this.working) {
					AsyncStreamsSSL.this.ba.setLastException(e);
					AsyncStreamsSSL.this.ba.raiseEventFromDifferentThread(
							AsyncStreamsSSL.this, null, 0,
							AsyncStreamsSSL.this.eventName + "_error", false,
							null);
				}
			}
		}

		private void closeUnexpected() throws IOException {
			AsyncStreamsSSL.this.ba
					.raiseEventFromDifferentThread(AsyncStreamsSSL.this, null,
							0, AsyncStreamsSSL.this.eventName + "_terminated",
							false, null);
			AsyncStreamsSSL.this.Close();
		}

		public void close() throws IOException {
			this.working = false;
			this.in.close();
		}
	}

	private class AOUT implements Runnable {
		private final OutputStream out;
		private volatile boolean working = true;
		private ArrayBlockingQueue<byte[]> queue = new ArrayBlockingQueue(100);
		private boolean prefix;
		private ByteBuffer bb;

		public AOUT(OutputStream out, boolean bigEndian, boolean prefix) {
			this.out = out;
			this.prefix = prefix;
			if (prefix) {
				this.bb = ByteBuffer.wrap(new byte[4]);
				this.bb.order(bigEndian ? ByteOrder.BIG_ENDIAN
						: ByteOrder.LITTLE_ENDIAN);
			}
		}

		public void run() {
			while (this.working)
				try {
					byte[] b = (byte[]) this.queue.take();
					this.out.write(b);
				} catch (Exception e) {
					if (this.working) {
						AsyncStreamsSSL.this.ba.setLastException(e);
						AsyncStreamsSSL.this.ba.raiseEventFromDifferentThread(
								AsyncStreamsSSL.this, null, 0,
								AsyncStreamsSSL.this.eventName + "_error",
								false, null);
					}
				}
		}

		public boolean put(byte[] buffer, int start, int len) {
			Log.w("B4A", "** Start AsyncStreamsSSL OutputStream **");
			byte[] b;
			if (!this.prefix) {
				b = new byte[len];
				System.arraycopy(buffer, start, b, 0, len);
			} else {
				b = new byte[len + 4];
				synchronized (this.bb) {
					this.bb.putInt(0, buffer.length);
					System.arraycopy(this.bb.array(), 0, b, 0, 4);
				}
				System.arraycopy(buffer, 0, b, 4, len);
			}
			try {
				Log.w("B4A", "** Offer AsyncStreamsSSL OutputStream **");
				return this.queue.offer(b, 100L, TimeUnit.MILLISECONDS);
			} catch (InterruptedException e) {
				Log.w("B4A", "** Error AsyncStreamsSSL OutputStream **");
			}
			throw new RuntimeException();
		}

		public void close() throws IOException {
			this.working = false;
			this.out.close();
		}
	}
}
